% Compute correspondence between nonmax suppressed pb and figure/ground
% boundaries.
% 
%    fg_pb_correspond(fg_data, radius, type)
%
% where type is 'train' or 'test' does the following:
%
%  - reads human f/g annotations according to fg_data
%  - reads stored pb from data/pres/
%  - assigns each corresponding nonmax suppressed pb point a figure/ground
%    label relative to its orientation (as defined by maximum pb response)
%    (points radius away on either side are used to determine the label)
%  - saves results in data/type/fg_pb_corr/
function fg_pb_correspond(fg_data, radius, type)
   % default radius
   if (nargin < 2), radius = 5; end
   % default type - training data
   if (nargin < 3), type = 'train'; end
   % create directory to store correspondence results
   out_dir = ['../data/' type '/fg_pb_corr'];
   if (~exist(out_dir,'dir')), mkdir(out_dir); end
   % set pb directory
   pb_dir = ['../data/' type '/pb_components'];
   % compute correspondence for each image
   n_images = numel(fg_data);
   for n = 1:n_images
      % read image
      iid = fg_data(n).iid;
      im = imgRead(iid,'color');
      % load first segmentation with f/g annotations
      ind = find(fg_data(n).fg_count > 0);
      if (isempty(ind)), continue; end
      seg_file = fg_data(n).seg_files{ind};
      fg_file  = fg_data(n).fg_files{ind}; fg_file = fg_file{1};
      seg = readSeg(seg_file);
      fg  = readbin(fg_file);
      % read pb components
      pb = load([pb_dir '/' num2str(iid) '.mat']);
      pb = mPb_all_from_cues_no_rescale( ...
         pb.bg1, pb.bg2, pb.bg3, ...
         pb.cga1, pb.cga2, pb.cga3, ...
         pb.cgb1, pb.cgb2, pb.cgb3, ...
         pb.tg1, pb.tg2, pb.tg3, 1.0);
      % nonmax suppress pb
      nmax = nonmax_channels(pb,pi/2);
      nmax = max(0, min(1, 1.2*nmax));
      edge = double(nmax > 0);
      % compute orientations
      [pb_max ori_ind] = max(pb,[],3);
      n_ori = size(pb,3);
      oris = (0:(n_ori-1))/n_ori * pi;
      % extract segment and contour labels
      bmap = double(seg2bmap(seg));
      [cmap,cid2sid] = seg2cmap(seg,bmap);
      fg_im     = fg{1};
      fg_labels = fg{2};
      % compute correspondance between pb and groundtruth contours
      [match1 match2] = correspondPixels(bmap, edge);
      % initialize figure/ground labeling for edge point
      [sx sy] = size(edge);
      edge_fg = zeros([sx sy]);
      % lookup figure/ground labeling for corresponded edge points
      inds = find(edge);
      for p = 1:numel(inds)
         % get center point and orientation
         ind = inds(p);
         [xc yc] = ind2sub([sx sy],ind);
         o = oris(ori_ind(ind))+ pi/2;
         % lookup matched point, skip if not matched
         ind_m = match2(ind);
         if (ind_m == 0), continue; end
         [xc_m yc_m] = ind2sub([sx sy],ind_m);
         % compute oriented offsets a distance radius away from matched point
         xa_m = round(xc_m + radius.*cos(o));
         ya_m = round(yc_m + radius.*sin(o));
         xb_m = round(xc_m - radius.*cos(o));
         yb_m = round(yc_m - radius.*sin(o));
         % skip if offsets are outside of image bounds
         if ((xa_m < 1) || (xa_m > sx) || (ya_m < 1) || (ya_m > sy) || ...
             (xb_m < 1) || (xb_m > sx) || (yb_m < 1) || (yb_m > sy) || ...
             (xc_m < 3) || (xc_m > sx-2) || (yc_m < 3) || (yc_m > sy-2))
            continue;
         end
         % lookup groundtruth segment labels
         label_a = seg(xa_m,ya_m);
         label_b = seg(xb_m,yb_m);
         % get labels of regions bordering groundtruth contour
         cid = cmap(ind_m);
         if (cid <= 0), continue; end
         r1 = cid2sid(1,cid);
         r2 = cid2sid(2,cid);
         % get figure/ground labels
         fg_local  = fg_im((xc_m-2):(xc_m+2),(yc_m-2):(yc_m+2));
         seg_local = seg((xc_m-2):(xc_m+2),(yc_m-2):(yc_m+2));
         r1_i = find(seg_local == r1);
         r2_i = find(seg_local == r2);
         if (~isempty(r1)), f1 = sign(mean(fg_local(r1_i))); else f1 = 0; end
         if (~isempty(r2)), f2 = sign(mean(fg_local(r2_i))); else f2 = 0; end
         % set f/g label for edge point
         if ((label_a == r1) && (label_b == r2))
            % f/g assignment at point is that of r1
            edge_fg(ind) = f1;
         elseif ((label_a == r2) && (label_b == r1))
            % f/g assignment at point is that of r2
            edge_fg(ind) = f2;
         else
            % region width less than radius, skip
            continue;
         end
      end
      % assemble results
      FG.pb      = pb;
      FG.nmax    = nmax;
      FG.ori_ind = ori_ind;
      FG.oris    = oris;
      FG.edge_fg = edge_fg;
      FG.radius  = radius;
      % save results
      save([out_dir '/' num2str(iid) '_fg_pb.mat'],'FG');
      % display results
      display_fg(FG);
      drawnow;
      %pause;
   end
end

function display_fg(FG)
   % extract fg
   [sx sy] = size(FG.edge_fg);
   inds = find(FG.edge_fg ~= 0);
   [xcs ycs] = ind2sub([sx sy],inds);
   os = (FG.oris(FG.ori_ind(inds)) + pi/2).';
   os = os + pi.*(FG.edge_fg(inds) < 0);
   xcs_a = xcs + (FG.radius).*cos(os);
   ycs_a = ycs + (FG.radius).*sin(os);
   % plot fg
   figure(1); imagesc(FG.nmax); axis image;
   figure(2); imagesc(FG.edge_fg); axis image;
   figure(3); imagesc(FG.nmax); axis image;
   title([num2str(numel(inds)) ' f/g training examples']);
   hold on;
   plot([ycs ycs_a].',[xcs xcs_a].','g');
   plot([ycs_a].',[xcs_a].','w.');
   hold off;
end
