% run simultaneous segmentation and figure/ground on images
% compute and save the angular embedding eigenvectors
function fg_run_eigs(fg_data, fg_classifier, type)
   % default type - training data
   if (nargin < 3), type = 'train'; end
   % set gb parameters
   rs      = [4 6 8 11];
   nthetas = [16 16 16 16];
   alpha   = 0.5;
   beta_gb = 1;
   % repulsion radius for gb
   rrep = 3;
   % maximum number of descriptors per image
   n_samp = 10.*1000;
   % threshold on for sampling nonmax suppressed pb
   nmax_thresh = 0.3;
   % connection radius
   radius = 5;
   % weight on ordering
   w_order = 0.0001;
   % set directories
   pb_dir   = ['../data/' type '/pb_components'];
   eigs_dir = ['../data/' type '/eigs'];
   % get number of images
   n_images = numel(fg_data);
   % predict segmentation and f/g labels
   for n = 1:n_images
      % read image
      iid = fg_data(n).iid;
      im = imgRead(iid,'color');
      % compute multiscale pb
      if (~exist([pb_dir '/' num2str(iid) '.mat'],'file'))
         %% compute cues - standard scales
         [bg1 bg2 bg3 cga1 cga2 cga3 cgb1 cgb2 cgb3 tg1 tg2 tg3 textons] = det_mPb(im);
         %% smooth cues
         tic;
         gtheta = [1.5708    1.1781    0.7854    0.3927   0    2.7489    2.3562    1.9635];
         for o = 1 : size(tg1, 3),
             bg1(:,:,o) = fitparab(bg1(:,:,o),3,3/4,gtheta(o));
             bg2(:,:,o) = fitparab(bg2(:,:,o),5,5/4,gtheta(o));
             bg3(:,:,o) = fitparab(bg3(:,:,o),10,10/4,gtheta(o));

             cga1(:,:,o) = fitparab(cga1(:,:,o),5,5/4,gtheta(o));
             cga2(:,:,o) = fitparab(cga2(:,:,o),10,10/4,gtheta(o));
             cga3(:,:,o) = fitparab(cga3(:,:,o),20,20/4,gtheta(o));

             cgb1(:,:,o) = fitparab(cgb1(:,:,o),5,5/4,gtheta(o));
             cgb2(:,:,o) = fitparab(cgb2(:,:,o),10,10/4,gtheta(o));
             cgb3(:,:,o) = fitparab(cgb3(:,:,o),20,20/4,gtheta(o));

             tg1(:,:,o) = fitparab(tg1(:,:,o),5,5/4,gtheta(o));
             tg2(:,:,o) = fitparab(tg2(:,:,o),10,10/4,gtheta(o));
             tg3(:,:,o) = fitparab(tg3(:,:,o),20,20/4,gtheta(o));
         end
         fprintf('Cues smoothing:%g\n', toc);
         % save result
         save([pb_dir '/' num2str(iid) '.mat'],'bg1','bg2','bg3','cga1','cga2','cga3','cgb1','cgb2','cgb3','tg1','tg2','tg3');
      end
      % read pb components
      pb = load([pb_dir '/' num2str(iid) '.mat']);
      pb_no_rescale = mPb_all_from_cues_no_rescale( ...
         pb.bg1, pb.bg2, pb.bg3, ...
         pb.cga1, pb.cga2, pb.cga3, ...
         pb.cgb1, pb.cgb2, pb.cgb3, ...
         pb.tg1, pb.tg2, pb.tg3, 1.0);
      pb = mPb_all_from_cues( ...
         pb.bg1, pb.bg2, pb.bg3, ...
         pb.cga1, pb.cga2, pb.cga3, ...
         pb.cgb1, pb.cgb2, pb.cgb3, ...
         pb.tg1, pb.tg2, pb.tg3, 1.0);
      [pbm ori] = max(pb,[],3);
      [sx sy] = size(pbm);
      % compute nonmax suppressed pb
      nmax = nonmax_channels(pb_no_rescale,pi/2);
      nmax = max(0, min(1, 1.2*nmax));
      % sample new descriptor locations
      [ci cj] = find_interest_points(nmax.*(nmax > nmax_thresh), n_samp, rrep);
      n_curr = numel(ci);
      inds = sub2ind([sx sy],ci,cj);
      % compute gb descriptors
      [descriptors sample_points] = ...
         compute_gb(nmax,ci,cj,alpha,beta_gb,rs,nthetas);
      % normalize descriptors
      descriptors = normalize_descriptors(descriptors,rs,nthetas);
      % rotate descriptors
      drot = zeros(size(descriptors));
      for idx = 1:n_curr
         frot = reshape(descriptors(idx,:),[16 4]);
         frot = circshift(frot,[(ori(inds(idx))-1) 0]);
         drot(idx,:) = reshape(frot,[1 64]);
      end
      % compute squared distance from each descriptor to each cluster center
      k = size(fg_classifier.shape_centroids,1);
      dmx = zeros([n_curr k]);
      for idx = 1:n_curr
         dist = repmat(drot(idx,:),[k 1]) - fg_classifier.shape_centroids;
         dist = sum(dist.*dist,2);
         dmx(idx,:) = dist;
      end
      % form feature matrix
      F = [exp(-dmx./(fg_classifier.sigma)) ones([n_curr 1])];
      % predict f/g
      preds = 1./(1+exp(-F*(fg_classifier.beta)));
      % show predictions
      [xcs ycs] = ind2sub([sx sy],inds);
      n_ori = 8;
      oris = (0:(n_ori-1))/n_ori * pi;
      os = (oris(ori(inds)) + pi/2).';
      os = os + pi.*(preds < 0.5);
      xcs_a = xcs + (radius).*abs(preds-0.5).*8.*cos(os).*nmax(inds);
      ycs_a = ycs + (radius).*abs(preds-0.5).*8.*sin(os).*nmax(inds);
      %figure(1); imagesc(nmax); axis image; colormap(gray);
      %hold on;
      %plot([ycs ycs_a].',[xcs xcs_a].','g');
      %plot([ycs_a].',[xcs_a].','w.');
      %hold off;
      %drawnow;
      % construct links
      xa = round(xcs + (radius).*cos(os));
      ya = round(ycs + (radius).*sin(os));
      xb = round(xcs - (radius).*cos(os));
      yb = round(ycs - (radius).*sin(os));
      g = find( ...
         (xa > 1) & (xa < sx) & (xb > 1) & (xb < sx) & ...
         (ya > 1) & (ya < sy) & (yb > 1) & (yb < sy));
      ia = sub2ind([sx sy],xa(g),ya(g));
      ib = sub2ind([sx sy],xb(g),yb(g));
      vals = repmat(pi./8,[numel(ia) 1]);
      conf = 2.*abs(preds(g)-0.5).*nmax(inds(g));
      index_a = cat(1, ia, ib);
      index_b = cat(1, ib, ia);
      vals_all = cat(1, vals, -vals);
      conf_all = cat(1, conf, conf);
      % construct affinity matrix
      rho = 0.1;
      A = pb_to_affinity(pb, rho);
      % create sparse ordering matrices
      [unique_idx u] = unique([index_a index_b],'rows');
      R   = sparse(index_a(u), index_b(u), vals_all(u), (sx*sy), (sx*sy));
      Cmx = sparse(index_a(u), index_b(u), conf_all(u), (sx*sy), (sx*sy));
      % compute eigenvectors for simulataneous grouping and f/g problem
      nvec = 16;
      if (~exist([eigs_dir '/' num2str(iid) '_eigs.mat'],'file'))
         [evecs_ae evals_ae] = ncut_ae(A + w_order.*Cmx, R, nvec);
         save([eigs_dir '/' num2str(iid) '_eigs.mat'],'evecs_ae','evals_ae');
      end
   end
end
