% compute a sparse affinity matrix from an oriented pb signal
function W = pb_to_affinity(pb_all, rho)
   % take maximum over orientations
   pb = max(pb_all,[],3);
   % interpolate pb on a large grid 2x original size
   [sx sy] = size(pb);
   sx_lg = (2*sx - 1);
   sy_lg = (2*sy - 1);
   pb_lg = imresize(pb, [sx_lg sy_lg], 'bilinear');
   % extract vertical neighbors
   x = repmat((1:(sx-1)).', [1 sy]);
   y = repmat((1:sy), [(sx-1) 1]);
   x_lg = 2.*x;
   y_lg = 2.*y - 1;
   i_vert_a = sub2ind([sx sy], x, y);
   i_vert_b = sub2ind([sx sy], x+1, y);
   val_vert = pb_lg(sub2ind([sx_lg sy_lg], x_lg, y_lg));
   % extract horizontal neighbors
   x = repmat((1:sx).', [1 (sy-1)]);
   y = repmat((1:(sy-1)), [sx 1]);
   x_lg = 2.*x - 1;
   y_lg = 2.*y;
   i_horz_a = sub2ind([sx sy], x, y);
   i_horz_b = sub2ind([sx sy], x, y+1);
   val_horz = pb_lg(sub2ind([sx_lg sy_lg], x_lg, y_lg));
   % extract upper-left lower-right diagonal neighbors
   x = repmat((1:(sx-1)).', [1 (sy-1)]);
   y = repmat((1:(sy-1)), [(sx-1) 1]);
   x_lg = 2.*x;
   y_lg = 2.*y;
   i_ullr_a = sub2ind([sx sy], x, y);
   i_ullr_b = sub2ind([sx sy], x+1, y+1);
   val_ullr = pb_lg(sub2ind([sx_lg sy_lg], x_lg, y_lg));
   % extract upper-right lower-left diagonal neighbors
   x = repmat((2:sx).', [1 (sy-1)]);
   y = repmat((1:(sy-1)), [(sx-1) 1]);
   x_lg = 2.*x - 2;
   y_lg = 2.*y;
   i_urll_a = sub2ind([sx sy], x, y);
   i_urll_b = sub2ind([sx sy], x-1, y+1);
   val_urll = pb_lg(sub2ind([sx_lg sy_lg], x_lg, y_lg));
   % extract self connections
   x = repmat((1:sx).', [1 sy]);
   y = repmat((1:sy), [sx 1]);
   i_self = sub2ind([sx sy], x, y);
   val_self = zeros(size(i_self));
   % get number of connections 
   n_vert = numel(val_vert);
   n_horz = numel(val_horz);
   n_ullr = numel(val_ullr);
   n_urll = numel(val_urll);
   n_self = numel(val_self);
   % reshape index and value vectors
   i_vert_a = reshape(i_vert_a, [n_vert 1]);
   i_vert_b = reshape(i_vert_b, [n_vert 1]);
   val_vert = reshape(val_vert, [n_vert 1]);
   i_horz_a = reshape(i_horz_a, [n_horz 1]);
   i_horz_b = reshape(i_horz_b, [n_horz 1]);
   val_horz = reshape(val_horz, [n_horz 1]);
   i_ullr_a = reshape(i_ullr_a, [n_ullr 1]);
   i_ullr_b = reshape(i_ullr_b, [n_ullr 1]);
   val_ullr = reshape(val_ullr, [n_ullr 1]);
   i_urll_a = reshape(i_urll_a, [n_urll 1]);
   i_urll_b = reshape(i_urll_b, [n_urll 1]);
   val_urll = reshape(val_urll, [n_urll 1]);
   i_self   = reshape(i_self,   [n_self 1]);
   val_self = reshape(val_self, [n_self 1]);
   % concatenate index and value vectors
   i_a = cat(1, ...
      i_vert_a, i_vert_b, ...
      i_horz_a, i_horz_b, ...
      i_ullr_a, i_ullr_b, ...
      i_urll_a, i_urll_b, ...
      i_self ...
   );
   i_b = cat(1, ...
      i_vert_b, i_vert_a, ...
      i_horz_b, i_horz_a, ...
      i_ullr_b, i_ullr_a, ...
      i_urll_b, i_urll_a, ...
      i_self ...
   );
   vals = cat(1, ...
      val_vert, val_vert, ...
      val_horz, val_horz, ...
      val_ullr, val_ullr, ...
      val_urll, val_urll, ...
      val_self ...
   );
   % convert pb values to affinity
   aff = exp(-vals./rho);
   % create sparse affinity matrix
   W = sparse(i_a, i_b, aff, (sx*sy), (sx*sy));
end
